/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */

package org.eobjects.metamodel.detect;

import java.math.BigInteger;
import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.eobjects.metamodel.schema.Column;
import org.eobjects.metamodel.schema.ColumnType;
import org.eobjects.metamodel.schema.MutableColumn;
import org.eobjects.metamodel.util.BooleanComparator;
import org.eobjects.metamodel.util.TimeComparator;

/**
 * Class used to detect and modify the column type based on incoming string
 * values. The life-cycle of ColumnTypeTransformer objects are as follows:
 * <ul>
 * <li>Construct with the column</li>
 * <li>Execute the registerValue(String) for all (distinct) values of the
 * column.</li>
 * <li>Execute the narrowColumnType() method once which sets the appropriate
 * column type on the column.</li>
 * <li>Execute the narrowValue(String) for all values that are queried before
 * returning them to the user.</li>
 * </ul>
 * The column type transformer is currently able to detect the following column
 * types:
 * <ul>
 * <li>Strings (default if no other type is found)</li>
 * <li>Booleans</li>
 * <li>Integers</li>
 * <li>Doubles</li>
 * <li>Date</li>
 * <li>Time</li>
 * </ul>
 * And it will also detect if the column is nullable, ie. if null occurs in the
 * values.
 */
public class ColumnTypeTransformer {

	private final static Logger logger = LoggerFactory
			.getLogger(ColumnTypeTransformer.class);

	private boolean _booleanPossible = true;
	private boolean _integerPossible = true;
	private boolean _doublePossible = true;
	private boolean _timestampPossible = true;
	private boolean _nullPossible = false;
	private MutableColumn _column;

	public ColumnTypeTransformer(MutableColumn column) {
		if (column == null) {
			throw new IllegalArgumentException("Column cannot be null");
		}
		_column = column;
	}

	public Column getColumn() {
		return _column;
	}

	public void registerValue(String valueAsString) {
		if (valueAsString == null) {
			_nullPossible = true;
		} else {
			if (_booleanPossible) {
				try {
					BooleanComparator.parseBoolean(valueAsString);
				} catch (IllegalArgumentException e) {
					_booleanPossible = false;
				}
			}
			if (_doublePossible) {
				try {
					Double.parseDouble(valueAsString);
				} catch (NumberFormatException e) {
					_doublePossible = false;
					_integerPossible = false;
				}
				// If integer is possible, double will always also be possible,
				// but not nescesarily the other way around
				if (_integerPossible) {
					try {
						Integer.parseInt(valueAsString);
					} catch (NumberFormatException e) {
						_integerPossible = false;
					}
				}
			}
			if (_timestampPossible) {
				if (TimeComparator.toDate(valueAsString) == null) {
					_timestampPossible = false;
				}
			}
		}
	}

	public void narrowColumnType() {
		logger.debug("narrowColumnType()");
		_column.setNullable(_nullPossible);
		if (_booleanPossible) {
			_column.setType(ColumnType.BOOLEAN);
			_column.setNativeType("Text transformed to boolean");
		} else if (_integerPossible) {
			_column.setType(ColumnType.INTEGER);
			_column.setNativeType("Text transformed to integer");
		} else if (_doublePossible) {
			_column.setType(ColumnType.DOUBLE);
			_column.setNativeType("Text transformed to double");
		} else if (_timestampPossible) {
			_column.setType(ColumnType.TIMESTAMP);
			_column.setNativeType("Text transformed to timestamp");
		} else {
			_column.setType(ColumnType.VARCHAR);
			_column.setNativeType("Text");
		}
		if (logger.isInfoEnabled()) {
			logger.info("Column type narrowed in column '" + _column.getName()
					+ "': type=" + _column.getType() + ", nullable="
					+ _column.isNullable());
		}
	}

	public Object narrowValue(String valueAsString) {
		if (valueAsString != null) {
			ColumnType type = _column.getType();
			switch (type) {
			case VARCHAR:
				return valueAsString;
			case BOOLEAN:
				return BooleanComparator.parseBoolean(valueAsString);
			case TINYINT:
			case SMALLINT:
				return Short.parseShort(valueAsString);
			case BIGINT:
				return new BigInteger(valueAsString);
			case INTEGER:
				return Integer.parseInt(valueAsString);
			case DECIMAL:
			case NUMERIC:
			case FLOAT:
			case REAL:
			case DOUBLE:
				return Double.parseDouble(valueAsString);
			case CHAR:
				if (valueAsString.length() == 1) {
					return valueAsString.charAt(0);
				} else {
					// This is kinda bad, but it's a fail over for some
					// datastores that actually DO store several chars in a CHAR
					// column
					return valueAsString;
				}
			case TIMESTAMP:
			case TIME:
			case DATE:
				Date date = TimeComparator.toDate(valueAsString);
				return date;
			default:
				throw new IllegalStateException(
						"Encountered unsupported column type: " + type);
			}
		}
		return null;
	}
}